#!/usr/bin/env python
# coding=utf-8

# aeneas is a Python/C library and a set of tools
# to automagically synchronize audio and text (aka forced alignment)
#
# Copyright (C) 2012-2013, Alberto Pettarin (www.albertopettarin.it)
# Copyright (C) 2013-2015, ReadBeyond Srl   (www.readbeyond.it)
# Copyright (C) 2015-2017, Alberto Pettarin (www.albertopettarin.it)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This module contains the following classes:

* :class:`~aeneas.hierarchytype.HierarchyType`,
  enumerating the allowed hierarchy types of a :class:`~aeneas.container.Container`.
"""


class HierarchyType:
    """
    Enumeration of the allowed hierarchy types of a
    :class:`~aeneas.container.Container`.
    """

    FLAT = "flat"
    """
    Flat hierarchy, that is, all the assets
    are located inside the same root directory
    (possibly, with subdirectories).
    """

    PAGED = "paged"
    """
    Paged hierarchy, that is, assets are divided
    into multiple sibling directories, one corresponding to each page/task.
    A page directory might have subdirectories where the audio/text assets
    are located.
    """

    ALLOWED_VALUES = [FLAT, PAGED]
    """ List of all the allowed values """
