#!/usr/bin/env python
# coding=utf-8

# aeneas is a Python/C library and a set of tools
# to automagically synchronize audio and text (aka forced alignment)
#
# Copyright (C) 2012-2013, Alberto Pettarin (www.albertopettarin.it)
# Copyright (C) 2013-2015, ReadBeyond Srl   (www.readbeyond.it)
# Copyright (C) 2015-2017, Alberto Pettarin (www.albertopettarin.it)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
aeneas.cwave is a Python C extension to read WAVE mono files.

.. function:: cwave.get_audio_info(audio_file_path)

    Read the sample rate and length of the given WAVE mono file.

    The returned tuple ``(sr, length)`` contains
    the sample rate and the number of samples
    of the WAVE file.

    :param string audio_file_path: the path of the WAVE file to be read, UTF-8 encoded
    :rtype: tuple

.. function:: cwave.read_audio_data(audio_file_path, from_sample, num_samples)

    Read audio samples from the given WAVE mono file.

    The returned tuple ``(sr, data)`` contains
    the sample rate of the WAVE file,
    and the samples read as a NumPy 1D array
    of ``float64`` values in ``[-1.0, 1.0]``.

    :param string audio_file_path: the path of the WAVE file to be read, UTF-8 encoded
    :param int from_sample: index of the first sample to be read
    :param int num_samples: number of samples to be read
    :rtype: tuple
"""
