 /************************************************************************/
 /*                                                                      */
 /*                Centre for Speech Technology Research                 */
 /*                     University of Edinburgh, UK                      */
 /*                       Copyright (c) 1996,1997                        */
 /*                        All Rights Reserved.                          */
 /*                                                                      */
 /*  Permission is hereby granted, free of charge, to use and distribute */
 /*  this software and its documentation without restriction, including  */
 /*  without limitation the rights to use, copy, modify, merge, publish, */
 /*  distribute, sublicense, and/or sell copies of this work, and to     */
 /*  permit persons to whom this work is furnished to do so, subject to  */
 /*  the following conditions:                                           */
 /*   1. The code must retain the above copyright notice, this list of   */
 /*      conditions and the following disclaimer.                        */
 /*   2. Any modifications must be clearly marked as such.               */
 /*   3. Original authors' names are not deleted.                        */
 /*   4. The authors' names are not used to endorse or promote products  */
 /*      derived from this software without specific prior written       */
 /*      permission.                                                     */
 /*                                                                      */
 /*  THE UNIVERSITY OF EDINBURGH AND THE CONTRIBUTORS TO THIS WORK       */
 /*  DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING     */
 /*  ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT  */
 /*  SHALL THE UNIVERSITY OF EDINBURGH NOR THE CONTRIBUTORS BE LIABLE    */
 /*  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES   */
 /*  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN  */
 /*  AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,         */
 /*  ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF      */
 /*  THIS SOFTWARE.                                                      */
 /*                                                                      */
 /*************************************************************************/
 /*                                                                       */
 /*                 Author: Richard Caley (rjc@cstr.ed.ac.uk)             */
 /* --------------------------------------------------------------------  */
 /* Uses preprocessor to define a template for asafe feature access       */
 /* function. This is just here to avoid massive duplication.             */
 /*                                                                       */
 /*************************************************************************/

#include "EST_features_aux.h"

#define defineGetFunction(FTYPE, VAL, RTYPE, NAME)   		\
RTYPE NAME(const FTYPE &it,					\
	   const EST_String name,				\
	   const RTYPE &def,					\
	   EST_feat_status &status)				\
{								\
  RTYPE val;							\
  EST_Val defv;							\
								\
  defv = est_val((void *)&defv);				\
								\
  CATCH_ERRORS()						\
    {								\
      if (strncmp(EST_error_message,"{FND}",5)==0)		\
        {							\
          status = efs_not_set;					\
          return def;						\
        }							\
								\
      status=efs_error;						\
      return def;						\
    }								\
								\
  EST_Val valr = it.VAL(name, defv);				\
								\
  if (valr.type() == val_type_pointer && &defv == pointer(valr))\
    {								\
    status=efs_not_set;						\
    val = def;							\
    }								\
  else								\
    {								\
    status=efs_ok;						\
    val = (RTYPE)valr;		       				\
    }								\
								\
  END_CATCH_ERRORS();						\
								\
  return val;							\
}




